@@ -32,38 +32,50 @@
 from __future__ import absolute_import
 
 import inspect
+import itertools
 import logging
 import os
 import platform
-import socket
+import re
 import sys
+import threading
 import time
+import urllib.parse
 import warnings
+
+import socket
+
 from copy import deepcopy
+from datetime import timedelta
+from typing import Optional, Sequence, Tuple
 
 import atatus
-from atatus.conf import Config, VersionedConfig, constants, update_config
-from atatus.conf.constants import ERROR, SPAN, TRANSACTION, METRICSET
+from atatus.conf import Config, VersionedConfig, constants
+from atatus.conf.constants import ERROR
 from atatus.metrics.base_metrics import MetricsRegistry
 from atatus.traces import Tracer, execution_context
-from atatus.utils import cgroup, compat, is_master_process, stacks, varmap
+from atatus.utils import cgroup, cloud, compat, is_master_process, stacks, varmap
+from atatus.utils.disttracing import TraceParent
 from atatus.utils.encoding import enforce_label_format, keyword_field, shorten, transform
 from atatus.utils.logging import get_logger
 from atatus.utils.module_import import import_string
-from atatus.utils.threading import IntervalTimer
+
 from atatus.collector.base import Collector
+from atatus.conf.constants import ERROR, SPAN, TRANSACTION, METRICSET
 from atatus.utils.hw_info import darwin_hwinfo, linux_hwinfo, default_hwinfo
 
 __all__ = ("Client",)
 
+CLIENT_SINGLETON = None
+
 
 class Client(object):
     """
-    The base Atatus APM client, which handles communication over the
+    The base Atatus client, which handles communication over the
     HTTP API to the APM Server.
 
     Will read default configuration from the environment variable
-    ``ATATUS_APP_NAME`` and ``ATATUS_LICENSE_KEY``
+    ``ATATUS_APP_NAME`` and ``ATATUS_SECRET_TOKEN``
     if available. ::
 
     >>> from atatus import Client
@@ -74,8 +86,8 @@ class Client(object):
     >>> # Configure the client manually
     >>> client = Client(
     >>>     include_paths=['my.package'],
-    >>>     app_name='myapp',
-    >>>     license_key='lic_apm_xxxxx',
+    >>>     service_name='myapp',
+    >>>     secret_token='secret_token',
     >>> )
 
     >>> # Record an exception
@@ -94,21 +106,36 @@ class Client(object):
         self.logger = get_logger("%s.%s" % (cls.__module__, cls.__name__))
         self.error_logger = get_logger("atatus.errors")
 
+        self._pid = None
+        self._thread_starter_lock = threading.Lock()
+        self._thread_managers = {}
+
         self.tracer = None
         self.processors = []
         self.filter_exception_types_dict = {}
         self._service_info = None
+        # setting server_version here is mainly used for testing
+        self.server_version = inline.pop("server_version", None)
+
+        self.check_python_version()
 
         config = Config(config, inline_dict=inline)
+        if config.deprecated_capture_body == "all":
+            config.capture_body = "all"
         if config.errors:
             for msg in config.errors.values():
                 self.error_logger.error(msg)
             config.disable_send = True
+        # if config.service_name == "unknown-python-service":
+        #     self.logger.warning("No custom SERVICE_NAME was set -- using non-descript default 'unknown-python-service'")
+        if config.service_name.strip() == "":
+            self.logger.error(
+                "SERVICE_NAME cannot be whitespace-only. Please set the SERVICE_NAME to a useful identifier."
+            )
         self.config = VersionedConfig(config, version=None)
 
         # Insert the log_record_factory into the logging library
-        # The LogRecordFactory functionality is only available on python 3.2+
-        if compat.PY3 and not self.config.disable_log_record_factory:
+        if not self.config.disable_log_record_factory:
             record_factory = logging.getLogRecordFactory()
             # Only way to know if it's wrapped is to create a log record
             throwaway_record = record_factory(__name__, logging.DEBUG, __file__, 252, "dummy_msg", [], None)
@@ -124,35 +151,35 @@ class Client(object):
         headers = {
             "Content-Type": "application/x-ndjson",
             "Content-Encoding": "gzip",
-            "User-Agent": "atatus-python/%s" % atatus.VERSION,
+            "User-Agent": self.get_user_agent(),
         }
 
-        # if self.config.secret_token:
-        #     headers["Authorization"] = "Bearer %s" % self.config.secret_token
         transport_kwargs = {
-            "metadata": self._build_metadata(),
+            "metadata": self.build_metadata(),
             "headers": headers,
             "verify_server_cert": self.config.verify_server_cert,
             "server_cert": self.config.server_cert,
             "timeout": self.config.server_timeout,
-            "max_flush_time": self.config.api_request_time / 1000.0,
-            "max_buffer_size": self.config.api_request_size,
+            "processors": self.load_processors(),
+            "analytics_processors": self.load_analytics_processors(),
         }
-        # self._api_endpoint_url = compat.urlparse.urljoin(
+
+        # self._api_endpoint_url = urllib.parse.urljoin(
         #     self.config.server_url if self.config.server_url.endswith("/") else self.config.server_url + "/",
         #     constants.EVENTS_API_PATH,
         # )
-        # self._transport = import_string(self.config.transport_class)(self._api_endpoint_url, **transport_kwargs)
+        # transport_class = import_string(self.config.transport_class)
+        # self._transport = transport_class(url=self._api_endpoint_url, client=self, **transport_kwargs)
+        # self.config.transport = self._transport
+        # self._thread_managers["transport"] = self._transport
 
-        self._collector = Collector(self.config, transport_kwargs["metadata"])
+        self._collector = Collector(self, transport_kwargs)
 
         for exc_to_filter in self.config.filter_exception_types or []:
             exc_to_filter_type = exc_to_filter.split(".")[-1]
             exc_to_filter_module = ".".join(exc_to_filter.split(".")[:-1])
             self.filter_exception_types_dict[exc_to_filter_type] = exc_to_filter_module
 
-        self.processors = [import_string(p) for p in self.config.processors] if self.config.processors else []
-
         if platform.python_implementation() == "PyPy":
             # PyPy introduces a `_functools.partial.__call__` frame due to our use
             # of `partial` in AbstractInstrumentedModule
@@ -176,6 +203,7 @@ class Client(object):
                         v,
                         list_length=self.config.local_var_list_max_length,
                         string_length=self.config.local_var_max_length,
+                        dict_length=self.config.local_var_dict_max_length,
                     ),
                     local_var,
                 ),
@@ -186,21 +214,40 @@ class Client(object):
         )
         self.include_paths_re = stacks.get_path_regex(self.config.include_paths) if self.config.include_paths else None
         self.exclude_paths_re = stacks.get_path_regex(self.config.exclude_paths) if self.config.exclude_paths else None
-        self._metrics = MetricsRegistry(
-            self.config.metrics_interval / 1000.0, self.queue, ignore_patterns=self.config.disable_metrics
-        )
+        self._metrics = MetricsRegistry(self)
         for path in self.config.metrics_sets:
             self._metrics.register(path)
         if self.config.breakdown_metrics:
             self._metrics.register("atatus.metrics.sets.breakdown.BreakdownMetricSet")
+        if self.config.prometheus_metrics:
+            self._metrics.register("atatus.metrics.sets.prometheus.PrometheusMetrics")
+        if self.config.metrics_interval:
+            self._thread_managers["metrics"] = self._metrics
         compat.atexit_register(self.close)
         if self.config.central_config:
-            self._config_updater = IntervalTimer(
-                update_config, 1, "eapm conf updater", daemon=True, args=(self,), evaluate_function_interval=True
-            )
-            self._config_updater.start()
+            self._thread_managers["config"] = self.config
         else:
             self._config_updater = None
+        if self.config.use_atatus_excepthook:
+            self.original_excepthook = sys.excepthook
+            sys.excepthook = self._excepthook
+        if config.enabled:
+            self.start_threads()
+
+        # Save this Client object as the global CLIENT_SINGLETON
+        set_client(self)
+
+    def start_threads(self):
+        current_pid = os.getpid()
+        if self._pid != current_pid:
+            with self._thread_starter_lock:
+                self.logger.debug("Detected PID change from %r to %r, starting threads", self._pid, current_pid)
+                for manager_type, manager in sorted(
+                    self._thread_managers.items(), key=lambda item: item[1].start_stop_order
+                ):
+                    self.logger.debug("Starting %s thread", manager_type)
+                    manager.start_thread(pid=current_pid)
+                self._pid = current_pid
 
     def get_handler(self, name):
         return import_string(name)
@@ -209,6 +256,8 @@ class Client(object):
         """
         Captures and processes an event and pipes it off to Client.send.
         """
+        if not self.config.is_recording:
+            return
         if event_type == "Exception":
             # never gather log stack for exceptions
             stack = False
@@ -248,19 +297,7 @@ class Client(object):
     def queue(self, event_type, data, flush=False):
         if self.config.disable_send:
             return
-        # Run the data through processors
-        for processor in self.processors:
-            if not hasattr(processor, "event_types") or event_type in processor.event_types:
-                data = processor(self, data)
-                if not data:
-                    self.logger.debug(
-                        "Dropped event of type %s due to processor %s.%s",
-                        event_type,
-                        getattr(processor, "__module__"),
-                        getattr(processor, "__name__"),
-                    )
-                    data = None  # normalize all "falsy" values to None
-                    break
+        self.start_threads()
         if flush and is_master_process():
             # don't flush in uWSGI master process to avoid ending up in an unpredictable threading state
             flush = False
@@ -276,16 +313,27 @@ class Client(object):
 
         # self._transport.queue(event_type, data, flush)
 
-    def begin_transaction(self, transaction_type, trace_parent=None, start=None):
+    def begin_transaction(
+        self,
+        transaction_type: str,
+        trace_parent: Optional[TraceParent] = None,
+        start: Optional[float] = None,
+        auto_activate: bool = True,
+        links: Optional[Sequence[TraceParent]] = None,
+    ):
         """
         Register the start of a transaction on the client
 
         :param transaction_type: type of the transaction, e.g. "request"
         :param trace_parent: an optional TraceParent object for distributed tracing
         :param start: override the start timestamp, mostly useful for testing
+        :param auto_activate: whether to set this transaction in execution_context
         :return: the started transaction object
         """
-        return self.tracer.begin_transaction(transaction_type, trace_parent=trace_parent, start=start)
+        if self.config.is_recording:
+            return self.tracer.begin_transaction(
+                transaction_type, trace_parent=trace_parent, start=start, auto_activate=auto_activate, links=links
+            )
 
     def end_transaction(self, name=None, result="", duration=None):
         """
@@ -296,16 +344,19 @@ class Client(object):
         :param duration: override duration, mostly useful for testing
         :return: the ended transaction object
         """
+        if duration is not None and not isinstance(duration, timedelta):
+            duration = timedelta(seconds=duration)
         transaction = self.tracer.end_transaction(result, name, duration=duration)
         return transaction
 
     def close(self):
-        if self._metrics:
-            self._metrics._stop_collect_timer()
-        if self._config_updater:
-            self._config_updater.cancel()
-        # self._transport.close()
+        if self.config.enabled:
+            with self._thread_starter_lock:
+                for _, manager in sorted(self._thread_managers.items(), key=lambda item: item[1].start_stop_order):
+                    manager.stop_thread()
         self._collector.close()
+        global CLIENT_SINGLETON
+        CLIENT_SINGLETON = None
 
     def get_service_info(self):
         if self._service_info:
@@ -331,6 +382,8 @@ class Client(object):
                 "name": keyword_field(self.config.framework_name),
                 "version": keyword_field(self.config.framework_version),
             }
+        if self.config.service_node_name:
+            result["node"] = {"configured_name": keyword_field(self.config.service_node_name)}
         self._service_info = result
         return result
 
@@ -344,7 +397,7 @@ class Client(object):
 
     def get_system_info(self):
         system_data = {
-            "hostname": keyword_field(socket.gethostname()),
+            "hostname": keyword_field(self.config.hostname),
             "architecture": platform.machine(),
             "platform": platform.system().lower(),
         }
@@ -371,6 +424,56 @@ class Client(object):
             system_data["kubernetes"] = k8s
         return system_data
 
+    def get_cloud_info(self):
+        """
+        Detects if the app is running in a cloud provider and fetches relevant
+        metadata from the cloud provider's metadata endpoint.
+        """
+        provider = str(self.config.cloud_provider).lower()
+
+        if not provider or provider == "none" or provider == "false":
+            return {}
+        if provider == "aws":
+            data = cloud.aws_metadata()
+            if not data:
+                self.logger.warning("Cloud provider {0} defined, but no metadata was found.".format(provider))
+            return data
+        elif provider == "gcp":
+            data = cloud.gcp_metadata()
+            if not data:
+                self.logger.warning("Cloud provider {0} defined, but no metadata was found.".format(provider))
+            return data
+        elif provider == "azure":
+            data = cloud.azure_metadata()
+            if not data:
+                self.logger.warning("Cloud provider {0} defined, but no metadata was found.".format(provider))
+            return data
+        elif provider == "auto" or provider == "true":
+            # Trial and error
+            data = {}
+            data = cloud.aws_metadata()
+            if data:
+                return data
+            data = cloud.gcp_metadata()
+            if data:
+                return data
+            data = cloud.azure_metadata()
+            return data
+        else:
+            self.logger.warning("Unknown value for CLOUD_PROVIDER, skipping cloud metadata: {}".format(provider))
+            return {}
+
+    def get_user_agent(self) -> str:
+        """
+        Compiles the user agent, which will be added as a header to all requests
+        to the APM Server
+        """
+        if self.config.service_version:
+            service_version = re.sub(r"[^\t _\x21-\x27\x2a-\x5b\x5d-\x7e\x80-\xff]", "_", self.config.service_version)
+            return "atatus-python-agent/{} ({} {})".format(atatus.VERSION, self.config.app_name, service_version)
+        else:
+            return "atatus-python-agent/{} ({})".format(atatus.VERSION, self.config.app_name)
+
     def get_hwinfo_info(self):
         if sys.platform.startswith('linux'):
             return linux_hwinfo()
@@ -379,16 +482,19 @@ class Client(object):
         else:
             return default_hwinfo()
 
-    def _build_metadata(self):
+    def build_metadata(self):
         data = {
             "service": self.get_service_info(),
             "process": self.get_process_info(),
             "system": self.get_system_info(),
+            "cloud": self.get_cloud_info(),
             "hwinfo": self.get_hwinfo_info(),
         }
         if "hwinfo" in data:
             if "bootid" not in data["hwinfo"]:
-                data["hwinfo"]["bootid"] = keyword_field(socket.gethostname())
+                data["hwinfo"]["bootid"] = keyword_field(socket.gethostname())        
+        if not data["cloud"]:
+            data.pop("cloud")
         if self.config.global_labels:
             data["labels"] = enforce_label_format(self.config.global_labels)
         return data
@@ -400,6 +506,7 @@ class Client(object):
         Captures, processes and serializes an event into a dict object
         """
         transaction = execution_context.get_transaction()
+        span = execution_context.get_span()
         if transaction:
             transaction_context = deepcopy(transaction.context)
         else:
@@ -422,6 +529,12 @@ class Client(object):
         event_data["context"] = context
         if transaction and transaction.labels:
             context["tags"] = deepcopy(transaction.labels)
+        # No intake for otel.attributes, so make them labels
+        if "otel_attributes" in context:
+            if context.get("tags"):
+                context["tags"].update(context.pop("otel_attributes"))
+            else:
+                context["tags"] = context.pop("otel_attributes")
 
         # if '.' not in event_type:
         # Assume it's a builtin
@@ -436,7 +549,7 @@ class Client(object):
         if custom.get("culprit"):
             culprit = custom.pop("culprit")
 
-        for k, v in compat.iteritems(result):
+        for k, v in result.items():
             if k not in event_data:
                 event_data[k] = v
 
@@ -458,6 +571,7 @@ class Client(object):
                         v,
                         list_length=self.config.local_var_list_max_length,
                         string_length=self.config.local_var_max_length,
+                        dict_length=self.config.local_var_dict_max_length,
                     ),
                     local_var,
                 ),
@@ -467,7 +581,7 @@ class Client(object):
         if "stacktrace" in log and not culprit:
             culprit = stacks.get_culprit(log["stacktrace"], self.config.include_paths, self.config.exclude_paths)
 
-        if "level" in log and isinstance(log["level"], compat.integer_types):
+        if "level" in log and isinstance(log["level"], int):
             log["level"] = logging.getLevelName(log["level"]).lower()
 
         if log:
@@ -491,11 +605,16 @@ class Client(object):
         if transaction:
             if transaction.trace_parent:
                 event_data["trace_id"] = transaction.trace_parent.trace_id
-            event_data["parent_id"] = transaction.id
+            # parent id might already be set in the handler
+            event_data.setdefault("parent_id", span.id if span else transaction.id)
             event_data["transaction_id"] = transaction.id
             if transaction.name:
                 event_data["event_transaction_name"] = transaction.name
-            event_data["transaction"] = {"sampled": transaction.is_sampled, "type": transaction.transaction_type}
+            event_data["transaction"] = {
+                "sampled": transaction.is_sampled,
+                "type": transaction.transaction_type,
+                "name": transaction.name,
+            }
 
         return event_data
 
@@ -516,7 +635,7 @@ class Client(object):
                     exc_name = "%s.%s" % (exc_module, exc_type)
                 else:
                     exc_name = exc_type
-                self.logger.info("Ignored %s exception due to exception type filter", exc_name)
+                self.logger.debug("Ignored %s exception due to exception type filter", exc_name)
                 return True
         return False
 
@@ -539,9 +658,99 @@ class Client(object):
             locals_processor_func=locals_processor_func,
         )
 
+    def _excepthook(self, type_, value, traceback):
+        try:
+            self.original_excepthook(type_, value, traceback)
+        except Exception:
+            self.capture_exception(handled=False)
+        finally:
+            self.capture_exception(exc_info=(type_, value, traceback), handled=False)
+
+    def load_processors(self):
+        """
+        Loads processors from self.config.processors, as well as constants.HARDCODED_PROCESSORS.
+        Duplicate processors (based on the path) will be discarded.
+
+        :return: a list of callables
+        """
+        processors = itertools.chain(self.config.processors, constants.HARDCODED_PROCESSORS)
+        seen = {}
+        # setdefault has the nice property that it returns the value that it just set on the dict
+        return [seen.setdefault(path, import_string(path)) for path in processors if path not in seen]
+
+
+    def load_analytics_processors(self):
+        """
+        Loads processors from self.config.analytics_processors.
+        Duplicate processors (based on the path) will be discarded.
+
+        :return: a list of callables
+        """
+        processors = itertools.chain(self.config.analytics_processors)
+        seen = {}
+        # setdefault has the nice property that it returns the value that it just set on the dict
+        return [seen.setdefault(path, import_string(path)) for path in processors if path not in seen]        
+
+    def should_ignore_url(self, url):
+        """Checks if URL should be ignored based on the transaction_ignore_urls setting"""
+        if self.config.transaction_ignore_urls:
+            for pattern in self.config.transaction_ignore_urls:
+                if pattern.match(url):
+                    return True
+        return False
+
+    def should_ignore_topic(self, topic: str) -> bool:
+        """Checks if messaging topic should be ignored based on the ignore_message_queues setting"""
+        if self.config.ignore_message_queues:
+            for pattern in self.config.ignore_message_queues:
+                if pattern.match(topic):
+                    return True
+        return False
+
+    def check_python_version(self):
+        v = tuple(map(int, platform.python_version_tuple()[:2]))
+        if v == (2, 7):
+            warnings.warn(
+                (
+                    "The Atatus agent will stop supporting Python 2.7 starting in 6.0.0 -- "
+                    "Please upgrade to Python 3.5+ to continue to use the latest features."
+                ),
+                PendingDeprecationWarning,
+            )
+        elif v < (3, 5):
+            warnings.warn("The Atatus agent only supports Python 3.5+", DeprecationWarning)
+
+    def check_server_version(
+        self, gte: Optional[Tuple[int, ...]] = None, lte: Optional[Tuple[int, ...]] = None
+    ) -> bool:
+        """
+        Check APM Server version against greater-or-equal and/or lower-or-equal limits, provided as tuples of integers.
+        If server_version is not set, always returns True.
+        :param gte: a tuple of ints describing the greater-or-equal limit, e.g. (7, 16)
+        :param lte: a tuple of ints describing the lower-or-equal limit, e.g. (7, 99)
+        :return: bool
+        """
+        if not self.server_version:
+            return True
+        gte = gte or (0,)
+        lte = lte or (2**32,)  # let's assume APM Server version will never be greater than 2^32
+        return bool(gte <= self.server_version <= lte)
+
 
 class DummyClient(Client):
     """Sends messages into an empty void"""
 
     def send(self, url, **kwargs):
         return None
+
+
+def get_client() -> Client:
+    return CLIENT_SINGLETON
+
+
+def set_client(client: Client):
+    global CLIENT_SINGLETON
+    if CLIENT_SINGLETON:
+        logger = get_logger("atatus")
+        logger.warning("Client object is being set more than once", stack_info=True)
+    CLIENT_SINGLETON = client
