@@ -4,7 +4,7 @@ import socket
 import platform
 from .layer import Layer
 import atatus
-
+from atatus.utils.stacks import get_lines_from_file
 
 class Builder(object):
     def __init__(self, config, metadata):
@@ -90,6 +90,13 @@ class Builder(object):
         payload["python"] = metric_data
         return payload
 
+    def analytics(self, start_time, end_time, analytics_data):
+        payload = self._common()
+        payload["startTime"] = start_time
+        payload["endTime"] = end_time
+        payload["requests"] = analytics_data
+        return payload
+
     def _build_hostinfo_obj(self):
         environment = {}
         environment["host"] = self._build_hostinfo_env_host_obj()
@@ -150,6 +157,9 @@ class Builder(object):
                     settings["frameworkVersion"] = self._metadata["service"]["framework"]["version"]
 
         settings["appName"] = self._config.app_name
+        settings["analytics"] = self._config.analytics
+        settings["logBody"] = self._config.capture_body
+        settings["traceThreshold"] = (self._config.trace_threshold.total_seconds() * 1000)
         settings["agentVersion"] = atatus.VERSION
 
         return settings
@@ -216,6 +226,31 @@ class Builder(object):
 
         return request
 
+    def _build_custom_data(self, context):
+        custom_data = {}
+        if "custom" in context and isinstance(context["custom"], dict):
+            for key, value in context["custom"].items():
+                try:
+                    custom_data[str(key)] = str(value)
+                except Exception as e:
+                    continue
+        
+        return custom_data
+
+    def _build_user(self, context):
+        user = {}
+        if "user" in context:
+            if "id" in context["user"]:
+                user["id"] = str(context["user"]["id"])
+
+            if "username" in context["user"]:
+                user["username"] = context["user"]["username"]
+
+            if "email" in context["user"]:
+                user["email"] = context["user"]["email"]
+
+        return user
+
     def _build_txns_obj(self, data):
         txns = []
         for t, v in data.items():
@@ -239,19 +274,28 @@ class Builder(object):
         traces = []
 
         for txn in traces_data:
-            if not all(k in txn for k in ("name", "timestamp", "duration")):
+            if not all(k in txn for k in ("name", "id", "timestamp", "duration")):
                 continue
 
             trace = {}
             trace["name"] = txn["name"]
             trace["type"] = self._config.framework_name or atatus.PYTHON_AGENT
             trace["kind"] = atatus.PYTHON_AGENT
-            trace["start"] = txn["timestamp"]
+            if "background" in txn:
+                trace["background"] = txn["background"]            
+            trace["start"] = (txn["timestamp"] // 1000)
             trace["duration"] = txn["duration"]
             if "context" in txn:
                 trace["request"] = self._build_request(txn["context"])
+                trace["customData"] = self._build_custom_data(txn["context"])
+                trace["user"] = self._build_user(txn["context"])
+
+            trace["txnId"] = txn["id"]
+            if "trace_id" in txn:
+                trace["traceId"] = txn["trace_id"]
             trace["entries"] = []
             trace["funcs"] = []
+
             i = 0
 
             if "spans" in txn:
@@ -270,15 +314,57 @@ class Builder(object):
                     entry["ly"]["name"] = span["name"]
                     entry["ly"]["type"] = Layer.types_dict.get(span["subtype"], span["subtype"])
                     entry["ly"]["kind"] = Layer.kinds_dict.get(span["type"], span["type"])
-                    if "context" in span:
+                    if "context" in span:                     
                         if "db" in span["context"]:
+                            if "dt" not in entry:
+                                entry["dt"] = {}
+                            entry["dt"]["db"] = {}                   
                             if "statement" in span["context"]["db"]:
-                                entry["dt"] = {}
-                                entry["dt"]["query"] = span["context"]["db"]["statement"]
+                                if span["context"]["db"]["statement"] is not None:
+                                    entry["dt"]["db"]["statement"] = span["context"]["db"]["statement"]
+                            if "type" in span["context"]["db"]:
+                                if span["context"]["db"]["type"] is not None:
+                                    entry["dt"]["db"]["type"] = span["context"]["db"]["type"]
+                            if "instance" in span["context"]["db"]:
+                                if span["context"]["db"]["instance"] is not None:
+                                    entry["dt"]["db"]["instance"] = span["context"]["db"]["instance"]
+                            if "user" in span["context"]["db"]:
+                                if span["context"]["db"]["user"] is not None:
+                                    entry["dt"]["db"]["user"] = span["context"]["db"]["user"]
+                            if "link" in span["context"]["db"]:
+                                if span["context"]["db"]["link"] is not None:
+                                    entry["dt"]["db"]["link"] = span["context"]["db"]["link"]
+                            if "rows_affected" in span["context"]["db"]:
+                                if span["context"]["db"]["rows_affected"] is not None:
+                                    entry["dt"]["db"]["rows_affected"] = span["context"]["db"]["rows_affected"]
+
                         if "http" in span["context"]:
-                            if "url" in span["context"]["http"]:
+                            if "dt" not in entry:
                                 entry["dt"] = {}
-                                entry["dt"]["url"] = span["context"]["http"]["url"]
+                            entry["dt"]["http"] = {}
+                            if "url" in span["context"]["http"]:
+                                if span["context"]["http"]["url"] is not None:
+                                    entry["dt"]["http"]["url"] = span["context"]["http"]["url"]
+                            if "status_code" in span["context"]["http"]:
+                                if span["context"]["http"]["status_code"] is not None:
+                                    entry["dt"]["http"]["status_code"] = span["context"]["http"]["status_code"]
+
+                        if "destination" in span["context"]:
+                            if "address" in span["context"]["destination"]:
+                                if span["context"]["destination"]["address"] is not None:
+                                    if "dt" not in entry:
+                                        entry["dt"] = {}
+                                    if "destination" not in entry["dt"]:
+                                        entry["dt"]["destination"] = {}
+                                    entry["dt"]["destination"]["address"] = span["context"]["destination"]["address"]
+                            if "port" in span["context"]["destination"]:                                        
+                                if span["context"]["destination"]["port"] is not None:
+                                    if "dt" not in entry:
+                                        entry["dt"] = {}                                    
+                                    if "destination" not in entry["dt"]:
+                                        entry["dt"]["destination"] = {}
+                                    entry["dt"]["destination"]["port"] = span["context"]["destination"]["port"]
+
                     try:
                         func_index = trace["funcs"].index(span["name"])
                     except ValueError:
@@ -327,14 +413,18 @@ class Builder(object):
         error_requests = []
 
         for v in requests_data:
-            if not all(k in v for k in ("name", "context")):
+            if not all(k in v for k in ("name", "timestamp", "context")):
                 continue
 
             error_request = {}
             error_request["name"] = v["name"]
             error_request["type"] = self._config.framework_name or atatus.PYTHON_AGENT
             error_request["kind"] = atatus.PYTHON_AGENT
+            error_request["timestamp"] = (v["timestamp"] // 1000)
             error_request["request"] = self._build_request(v["context"])
+            error_request["customData"]  = self._build_custom_data(v["context"])
+            error_request["user"]  = self._build_user(v["context"])
+
             error_requests.append(error_request)
 
         return error_requests
@@ -342,25 +432,50 @@ class Builder(object):
     def _build_errors_obj(self, errors_data):
         errors = []
         for v in errors_data:
-            if not all(k in v for k in ("timestamp", "exception")):
+            if "timestamp" not in v:
                 continue
 
-            if not all(k in v["exception"] for k in ("type", "message")):
+            errorObj = {}
+            if "exception" in v:
+                errorObj = v["exception"]
+            elif "log" in v:
+                errorObj = v["log"]
+            else:
+                continue
+
+            if "message" not in errorObj:
                 continue
 
             error = {}
-            error["timestamp"] = v["timestamp"]
+            error["timestamp"] = (v["timestamp"] // 1000)
+
             if "event_transaction_name" in v:
                 error["transaction"] = v["event_transaction_name"]
+
+            if "transaction_id" in v:
+                error["txnId"] = v["transaction_id"]
+
+            if "trace_id" in v:
+                error["traceId"] = v["trace_id"]
+
             if "context" in v:
                 error["request"] = self._build_request(v["context"])
+                error["customData"]  = self._build_custom_data(v["context"])
+                error["user"]  = self._build_user(v["context"])
+
             error["exceptions"] = []
             exception = {}
-            exception["class"] = v["exception"]["type"]
-            exception["message"] = v["exception"]["message"]
+            if "type" in errorObj:
+                exception["class"] = errorObj["type"]
+            elif "level" in errorObj:
+                exception["class"] = errorObj["level"]
+            else:
+                exception["class"] = "Unknown"
+
+            exception["message"] = errorObj["message"]
             exception["stacktrace"] = []
-            if "stacktrace" in v["exception"]:
-                for f in v["exception"]["stacktrace"]:
+            if "stacktrace" in errorObj:
+                for f in errorObj["stacktrace"]:
                     if not all(k in f for k in ("filename", "function", "lineno", "library_frame")):
                         continue
 
@@ -371,26 +486,36 @@ class Builder(object):
                     if f["library_frame"] is False:
                         frame["inp"] = True
 
-                    if "context_line" in f:
-                        frame["code"] = []
+                    if "context_metadata" in f:
+                        fname, lineno, context_lines, loader, module_name = f.pop("context_metadata")
+                        pre_context, context_line, post_context = get_lines_from_file(
+                            fname, lineno, context_lines, loader, module_name
+                        )
+                        if context_line:
+                            f["pre_context"] = pre_context
+                            f["context_line"] = context_line
+                            f["post_context"] = post_context
 
-                        if "pre_context" in f:
-                            psize = len(f["pre_context"])
-                            lineno = 0
-                            if f["lineno"] - psize > 0:
-                                lineno = f["lineno"] - psize
-                            for c in f["pre_context"]:
-                                frame["code"].append([str(lineno), c])
-                                lineno += 1
+                        if "context_line" in f:
+                            frame["code"] = []
 
-                        frame["code"].append([str(f["lineno"]), f["context_line"]])
+                            if "pre_context" in f:
+                                psize = len(f["pre_context"])
+                                lineno = 0
+                                if f["lineno"] - psize > 0:
+                                    lineno = f["lineno"] - psize
+                                for c in f["pre_context"]:
+                                    frame["code"].append([str(lineno), c])
+                                    lineno += 1
 
-                        if "post_context" in f:
-                            psize = len(f["post_context"])
-                            lineno = f["lineno"] + 1
-                            for c in f["post_context"]:
-                                frame["code"].append([str(lineno), c])
-                                lineno += 1
+                            frame["code"].append([str(f["lineno"]), f["context_line"]])
+
+                            if "post_context" in f:
+                                psize = len(f["post_context"])
+                                lineno = f["lineno"] + 1
+                                for c in f["post_context"]:
+                                    frame["code"].append([str(lineno), c])
+                                    lineno += 1
 
                     exception["stacktrace"].append(frame)
 
